import json
import structlog
from sqlalchemy import orm, engine
from sqlalchemy import and_, or_

from opentaxii.persistence import OpenTAXIIPersistenceAPI

from . import models
from . import converters as conv

__all__ = ['SQLDatabaseAPI']

log = structlog.getLogger(__name__)


class SQLDatabaseAPI(OpenTAXIIPersistenceAPI):
    """SQL database implementation of OpenTAXII Persistence API.

    Implementation will work with any DB supported by SQLAlchemy package.

    :param str db_connection: a string that indicates database dialect and
                          connection arguments that will be passed directly
                          to :func:`~sqlalchemy.engine.create_engine` method.

    :param bool create_tables=False: if True, tables will be created in the DB.
    """

    def __init__(self, db_connection, create_tables=False):

        self.engine = engine.create_engine(db_connection, convert_unicode=True)

        self.Session = orm.scoped_session(orm.sessionmaker(autocommit=False,
            autoflush=True, bind=self.engine))

        attach_all(self, models)

        self.Base.query = self.Session.query_property()

        if create_tables:
            self.create_tables()


    def create_tables(self):
        self.Base.metadata.create_all(bind=self.engine)


    def _merge(self, obj):
        s = self.Session()
        updated = s.merge(obj)
        s.commit()
        return updated


    def get_services(self, collection_id=None):

        query = self.Service.query

        if collection_id:
            query = query.filter(self.Service.collections.any(id=collection_id))

        return map(conv.to_service_entity, query.all())
 

    def get_service(self, sid):
        s = self.Service.get(sid)
        return conv.to_service_entity(s)


    def update_service(self, entity):
        service = self.Service(id=entity.id, type=entity.type,
                properties=entity.properties)
        updated = self._merge(service)
        return conv.to_service_entity(updated)


    def create_service(self, entity):
        return self.update_service(entity)


    def get_collections(self, service_id):
        service = self.Service.query.get(service_id)
        return map(conv.to_collection_entity, service.collections)


    def get_collection(self, name, service_id):

        service = self.Service.query.get(service_id)

        for c in service.collections:
            if c.name == name:
                return conv.to_collection_entity(c)


    def _get_content_query(self, collection_id=None, start_time=None,
            end_time=None, bindings=None):

        query = self.ContentBlock.query

        if collection_id:
            query = query.filter(self.ContentBlock.collections.any(id=collection_id))

        if start_time:
            query = query.filter(self.ContentBlock.date_created > start_time)

        if end_time:
            query = query.filter(self.ContentBlock.date_created <= end_time)

        if bindings:
            criteria = []
            for binding in bindings:
                if binding.subtypes:
                    criterion = and_(
                            self.ContentBlock.binding_id == binding.binding,
                            self.ContentBlock.binding_subtype.in_(binding.subtypes)
                            )
                else:
                    criterion = self.ContentBlock.binding_id == binding.binding
                criteria.append(criterion)

            query = query.filter(or_(*criteria))

        return query


    def get_content_blocks_count(self, collection_id=None, start_time=None,
            end_time=None, bindings=None):

        query = self._get_content_query(collection_id=collection_id,
                start_time=start_time, end_time=end_time,
                bindings=bindings)

        return query.count()


    def get_content_blocks(self, collection_id=None, start_time=None,
            end_time=None, bindings=None, offset=0, limit=10):

        query = self._get_content_query(collection_id=collection_id,
                start_time=start_time, end_time=end_time,
                bindings=bindings)
        
        blocks = query[offset : offset + limit]

        return map(conv.to_block_entity, blocks)


    def update_collection(self, entity):

        _bindings = conv.serialize_content_bindings(entity.supported_content)

        collection = self.DataCollection(
            id = entity.id,
            name = entity.name,
            type = entity.type,
            description = entity.description,
            available = entity.available,
            accept_all_content = entity.accept_all_content,
            bindings = _bindings
        )

        updated = self._merge(collection)

        log.debug("Collection updated", collection_id=updated.id,
                collection_name=updated.name)

        return conv.to_collection_entity(updated)


    def create_collection(self, entity):
        return self.update_collection(entity)


    def attach_collection_to_services(self, collection_id, service_ids):
        
        collection = self.DataCollection.query.get(collection_id)

        if not collection:
            raise ValueError("Can't find collection with id=%s" % collection_id)

        s = self.Session()

        for sid in service_ids:
            service = self.Service.query.get(sid)
            collection.services.append(service)

        s.add(collection)
        s.commit()

        log.debug("Collection attached", collection_id=collection.id,
                collection_name=collection.name, service_ids=service_ids)


    def create_inbox_message(self, entity):
        return self.update_inbox_message(entity)


    def update_inbox_message(self, entity):

        if entity.destination_collections:
            names = json.dumps(entity.destination_collections) 
        else:
            names = None

        message = self.InboxMessage(
            id = entity.id,
            message_id = entity.message_id,
            original_message = entity.original_message,
            content_block_count = entity.content_block_count,
            destination_collections = names,

            service_id = entity.service_id,

            result_id = entity.result_id,
            record_count = entity.record_count,
            partial_count = entity.partial_count,

            subscription_collection_name = entity.subscription_collection_name,
            subscription_id = entity.subscription_id,

            exclusive_begin_timestamp_label = entity.exclusive_begin_timestamp_label,
            inclusive_end_timestamp_label = entity.inclusive_end_timestamp_label,
        )

        updated = self._merge(message)

        return conv.to_inbox_message_entity(updated)


    def update_content_block(self, entity):

        if entity.content_binding:
            binding = entity.content_binding.binding
            subtype = entity.content_binding.subtypes[0] \
                    if entity.content_binding.subtypes else None
        else:
            binding = None
            subtype = None

        content = self.ContentBlock(
            id = entity.id,
            timestamp_label = entity.timestamp_label,
            inbox_message_id = entity.inbox_message_id,
            content = entity.content,
            binding_id = binding,
            binding_subtype = subtype
        )

        updated = self._merge(content)

        return conv.to_block_entity(updated)


    def create_content_block(self, entity, collection_ids=None,
            service_id=None):
        block = self.update_content_block(entity)

        if collection_ids:
            self._attach_content_to_collections(block, collection_ids)

        return block


    def _attach_content_to_collections(self, content_block, collection_ids):

        if not collection_ids:
            return

        content_block_id = content_block.id

        s = self.Session()
        content_block = self.ContentBlock.query.get(content_block_id)

        criteria = self.DataCollection.id.in_(collection_ids)
        collections = self.DataCollection.query.filter(criteria).all()

        if not collections:
            raise ValueError("No collections were found with ids: %s" % collection_ids)

        for collection in collections:
            collection.content_blocks.append(content_block)
            s.add(collection)

            log.debug("Content block added to collection",
                    content_block_id=content_block_id,
                    collection_id=collection.id,
                    collection_name=collection.name)

        s.commit()



    def update_result_set(self, entity):

        _bindings = conv.serialize_content_bindings(entity.content_bindings)

        result_set = self.ResultSet(
            id = entity.result_id,
            collection_id = entity.collection_id,
            bindings = _bindings,
            begin_time = entity.timeframe[0],
            end_time = entity.timeframe[1]
        )

        updated = self._merge(result_set)

        return conv.to_result_set_entity(updated)

    def create_result_set(self, entity):
        return self.update_result_set(entity)

    def get_result_set(self, result_set_id):
        result_set = self.ResultSet.query.get(result_set_id)
        return conv.to_result_set_entity(result_set)

    def get_subscription(self, subscription_id):
        s = self.Subscription.query.get(subscription_id)
        return conv.to_subscription_entity(s)

    def get_subscriptions(self, service_id):
        service = self.Service.query.get(service_id)
        return map(conv.to_subscription_entity, service.subscriptions)

    def update_subscription(self, entity):

        if entity.params:
            params = dict(
                response_type = entity.params.response_type,
                content_bindings = conv.serialize_content_bindings(
                        entity.params.content_bindings)
            )
        else:
            params = {}

        subscription = self.Subscription(
            id = entity.subscription_id,
            collection_id = entity.collection_id,
            params = json.dumps(params),
            status = entity.status,
            service_id = entity.service_id
        )

        updated = self._merge(subscription)

        log.debug("Subscription updated", subscription_id=updated.id,
                collection_id=updated.collection_id, status=updated.status)

        return conv.to_subscription_entity(updated)

    def create_subscription(self, entity):
        return self.update_subscription(entity)


def attach_all(obj, module):
    for key in module.__all__:
        if not hasattr(obj, key):
            setattr(obj, key, getattr(module, key))
    return obj





import structlog

from .taxii.services import (
        DiscoveryService, InboxService, CollectionManagementService,
        PollService
)
from .config import ServerConfig
from .persistence import PersistenceManager
from .auth import AuthManager
from .utils import get_path_and_address, attach_signal_hooks, load_api

log = structlog.get_logger(__name__)

class TAXIIServer(object):
    '''TAXII Server class.
    
    This class keeps Presistence API and Auth API managers instances
    and creates TAXII Service instances on request.

    This class can be initiated directly but should be created via
    :py:func:`create_server`.

    :param str domain: domain name that will be used in
                       services absolute address values
    :param `opentaxii.persistence.manager.PersistenceManager` persistence_manager:
                       persistence manager instance
    :param `opentaxii.auth.manager.AuthManager` auth_manager:
                       authentication manager instance
    '''

    TYPE_TO_SERVICE = dict(
        inbox = InboxService,
        discovery = DiscoveryService,
        collection_management = CollectionManagementService,
        poll = PollService
    )

    def __init__(self, domain, persistence_manager, auth_manager):

        self.domain = domain

        self.persistence = persistence_manager
        self.auth = auth_manager


    def _create_services(self, service_entities):

        discovery_services = []
        services = []

        for entity in service_entities:

            _props = dict(entity.properties)
            _props['server'] = self

            _props['path'], _props['address'] = get_path_and_address(
                    self.domain, _props['address'])

            advertised = _props.pop('advertised_services', None)

            if entity.type not in self.TYPE_TO_SERVICE:
                raise ValueError('Unknown service type "%s"' % entity.type)

            service = self.TYPE_TO_SERVICE[entity.type](id=entity.id, **_props)

            services.append(service)

            if advertised:
                discovery_services.append((service, advertised))

        for service, advertised in discovery_services:
            service.set_advertised_services([s for s in services if s.id in advertised])

        return services

    def get_services(self, ids=None):
        '''Get services registered with this TAXII server instance.

        :param list ids: list of service IDs (as strings) to use as a filter

        :return: list of services
        :rtype: list of :py:class:`opentaxii.taxii.services.abstract.TAXIIService`
        '''

        if ids is not None and len(ids) == 0:
            return []

        service_entities = self.persistence.get_services()
        # Services needs to be created all at once to ensure that
        # discovery services list all active advertised services
        services = self._create_services(service_entities)

        if ids:
            services = filter(lambda s: s.id in ids, services)

        return services

    def get_service(self, id):
        '''Get service by ID.

        :param str id: service ID

        :return: service with specified ID or None
        :rtype: :py:class:`opentaxii.taxii.services.abstract.TAXIIService`
        '''
        
        services = self.get_services([id])
        if services:
            return services[0]

    def get_services_for_collection(self, collection, service_type):
        '''Get list of services with type ``service_type``, attached
        to collection ``collection``.

        :param `opentaxii.taxii.entities.CollectionEntity` collection:
                    collection in question
        :param str service_type: service type, supported values are
                    listed as keys in :py:attr:`TYPE_TO_SERVICE`

        :return: list of services
        :rtype: list of :py:class:`opentaxii.taxii.services.abstract.TAXIIService`
        '''

        if service_type not in self.TYPE_TO_SERVICE:
            raise ValueError('Wrong service type: %s' % service_type)

        entities = self.persistence.get_services_for_collection(collection)

        requested_entities = [e.id for e in entities if e.type == service_type]

        if not requested_entities:
            return []

        # Sync services for collection with registered services for this server
        services = self.get_services(requested_entities)

        return services


def create_server(config=None):
    '''Create TAXII server instance.

    Function gets required parameters from the config (domain name,
    API class names, hooks module name), creates API instances and configures
    new TAXII server instance.

    :param `opentaxii.config.ServerConfig` config: config instance

    :return: TAXII server instance
    :rtype: :py:class:`opentaxii.server.TAXIIServer`
    '''

    config = config or ServerConfig()
    attach_signal_hooks(config)

    persistence_api = load_api(config['persistence_api'])
    persistence_manager = PersistenceManager(api=persistence_api)

    auth_api = load_api(config['auth_api'])
    auth_manager = AuthManager(api=auth_api)

    domain = config['domain']
    server = TAXIIServer(domain, persistence_manager=persistence_manager,
            auth_manager=auth_manager)

    return server




import structlog
from flask import Flask, request, make_response, abort

from .taxii.exceptions import (
    raise_failure, StatusMessageException, FailureStatus, UnauthorizedStatus
)
from .taxii.utils import parse_message
from .taxii.status import process_status_exception
from .taxii.bindings import (
    MESSAGE_BINDINGS, SERVICE_BINDINGS, ALL_PROTOCOL_BINDINGS
)
from .taxii.http import (
    get_http_headers, get_content_type, validate_request_headers_post_parse,
    validate_request_headers, validate_response_headers
)
from .utils import extract_token
from .management import management

log = structlog.get_logger(__name__)


def create_app(server):
    '''Create Flask application and attach TAXII server
    instance ``server`` to it.

    :param `opentaxii.server.TAXIIServer` server: TAXII server instance

    :return: Flask app
    '''

    app = Flask(__name__)

    app.taxii = server

    app.add_url_rule("/<path:relative_path>", "opentaxii_services_view",
            _server_wrapper(server), methods=['POST'])

    app.register_blueprint(management, url_prefix='/management')

    app.register_error_handler(500, handle_internal_error)
    app.register_error_handler(StatusMessageException, handle_status_exception)

    return app


def _server_wrapper(server):

    def wrapper(relative_path=""):

        relative_path = '/' + relative_path

        for service in server.get_services():
            if service.path and service.path == relative_path:
                return _process_with_service(service)

        abort(404)

    return wrapper


def _process_with_service(service):

    if not service.available:
        raise_failure("The service is not available")

    if service.authentication_required:
        token = extract_token(request.headers)
        if not token:
            raise UnauthorizedStatus()

        account = service.server.auth.get_account(token)
        if not account:
            raise UnauthorizedStatus()

    if 'application/xml' not in request.accept_mimetypes:
        raise_failure("The specified values of Accept is not supported: %s" %
                ", ".join((request.accept_mimetypes or [])))

    validate_request_headers(request.headers, MESSAGE_BINDINGS)

    taxii_message = parse_message(get_content_type(request.headers), request.data)

    try:
        validate_request_headers_post_parse(request.headers,
                supported_message_bindings=MESSAGE_BINDINGS,
                service_bindings=SERVICE_BINDINGS,
                protocol_bindings=ALL_PROTOCOL_BINDINGS)
    except StatusMessageException, e:
        e.in_response_to = taxii_message.message_id
        raise e

    response_message = service.process(request.headers, taxii_message)

    response_headers = get_http_headers(response_message.version, request.is_secure)
    validate_response_headers(response_headers)

    # FIXME: pretty-printing should be configurable
    taxii_xml = response_message.to_xml(pretty_print=True)

    return make_taxii_response(taxii_xml, response_headers)


def make_taxii_response(taxii_xml, taxii_headers):

    validate_response_headers(taxii_headers)
    response = make_response(taxii_xml)

    h = response.headers
    for header, value in taxii_headers.items():
        h[header] = value

    return response


def handle_status_exception(error):
    log.warning('Status exception', exc_info=True)

    if 'application/xml' not in request.accept_mimetypes:
        return 'Unacceptable', 406

    xml, headers = process_status_exception(error, request.headers, request.is_secure)
    return make_taxii_response(xml, headers)


def handle_internal_error(error):
    log.error('Internal error', exc_info=True)

    if 'application/xml' not in request.accept_mimetypes:
        return 'Unacceptable', 406

    new_error = FailureStatus("Error occured", e=error)

    xml, headers = process_status_exception(new_error, request.headers, request.is_secure)
    return make_taxii_response(xml, headers)





class Account(object):
    '''Represents Account entity.

    This class holds user-specific information and is used
    for authorization.

    :param str id: account id
    '''

    def __init__(self, id):
        self.id = id
    



import structlog

import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10
from libtaxii.constants import (
    SD_SUPPORTED_CONTENT, ST_UNSUPPORTED_CONTENT_BINDING,
    SD_ITEM, ST_NOT_FOUND, ST_BAD_MESSAGE,
    ACT_SUBSCRIBE, ACT_UNSUBSCRIBE, ACT_PAUSE,
    ACT_RESUME, ACT_STATUS, ACT_TYPES_11,
    ACT_TYPES_10
)

from ...exceptions import StatusMessageException, raise_failure
from ...converters import subscription_to_subscription_instance, parse_content_bindings
from ...entities import PollRequestParametersEntity, SubscriptionEntity

from .base_handlers import BaseMessageHandler
from .poll_request_handlers import retrieve_collection

log = structlog.getLogger(__name__)


def action_subscribe(request, service, collection, version, **kwargs):

    if version == 11:
        params = request.subscription_parameters
        response_type = params.response_type

        if len(params.content_bindings) == 0:
            supported_contents = []
        else:
            requested_bindings = parse_content_bindings(params.content_bindings, version=version)
            supported_contents = collection.get_matching_bindings(requested_bindings)

            if requested_bindings and not supported_contents:
                details = {SD_SUPPORTED_CONTENT: collection.get_supported_content(version=version)}
                raise StatusMessageException(ST_UNSUPPORTED_CONTENT_BINDING,
                        in_response_to=request.message_id, status_details=details)

    else:
        supported_contents = []
        response_type = None

    poll_request_params = PollRequestParametersEntity(
        response_type = response_type,
        content_bindings = supported_contents,
    )

    # we are ignoring Delivery Parameters for now

    subscription = SubscriptionEntity(
        service_id = service.id,
        collection_id = collection.id,
        poll_request_params = poll_request_params,
        status = SubscriptionEntity.ACTIVE
    )

    return service.create_subscription(subscription)


def action_unsubscribe(request, service, subscription, **kwargs):
    if subscription:
        subscription.status = SubscriptionEntity.UNSUBSCRIBED
        return service.update_subscription(subscription)
    else:
        return SubscriptionEntity(collection_id=None, service_id=service.id,
                subscription_id=request.subscription_id,
                status=SubscriptionEntity.UNSUBSCRIBED)


def action_status(service, subscription, **kwargs):
    if subscription:
        return subscription
    else:
        return service.get_subscriptions()


def action_pause(service, subscription, **kwargs):
    if subscription.status == SubscriptionEntity.PAUSED:
        return subscription
    subscription.status = SubscriptionEntity.PAUSED
    return service.update_subscription(subscription)

def action_resume(service, subscription, **kwargs):
    if subscription.status != SubscriptionEntity.PAUSED:
        return subscription
    subscription.status = SubscriptionEntity.ACTIVE
    return service.update_subscription(subscription)


ACTIONS = {
    ACT_SUBSCRIBE: action_subscribe,
    ACT_UNSUBSCRIBE: action_unsubscribe,
    ACT_PAUSE: action_pause,
    ACT_RESUME: action_resume,
    ACT_STATUS: action_status
}


class SubscriptionRequest11Handler(BaseMessageHandler):

    supported_request_messages = [tm11.ManageCollectionSubscriptionRequest]

    @classmethod
    def validate_request(cls, request, subscription):

        action = request.action

        if action not in ACT_TYPES_11:
            error_message = "The specified action was invalid"

        elif action in (ACT_UNSUBSCRIBE, ACT_PAUSE, ACT_RESUME) \
                and not request.subscription_id:
            error_message = 'Action "%s" requires a subscription id' % action

        else:
            error_message = None

        if error_message:
            raise StatusMessageException(ST_BAD_MESSAGE, message=error_message,
                    in_response_to=request.message_id)

        if not subscription and (action in (ACT_PAUSE, ACT_RESUME) or 
                (action == ACT_STATUS and request.subscription_id)):

            details = {SD_ITEM: request.subscription_id}
            raise StatusMessageException(ST_NOT_FOUND, status_details=details,
                    in_response_to=request.message_id)


    @classmethod
    def handle_message(cls, service, request):

        if request.subscription_id:
            subscription = service.get_subscription(request.subscription_id)
        else:
            subscription = None

        cls.validate_request(request, subscription)

        collection = retrieve_collection(service, request.collection_name, request.message_id)

        if subscription and subscription.collection_id != collection.id:
            details = {SD_ITEM: request.collection_name}
            raise StatusMessageException(ST_NOT_FOUND, status_details=details,
                    in_response_to=request.message_id)


        response = tm11.ManageCollectionSubscriptionResponse(
            message_id = cls.generate_id(),
            in_response_to = request.message_id,
            collection_name = collection.name,
            message = service.subscription_message,
        )

        result = ACTIONS[request.action](service=service, request=request,
                collection=collection, subscription=subscription, version=11)

        if isinstance(result, list):
            results = result
        else:
            results = [result]

        polling_services = service.get_polling_services(collection)

        for s in results:
            instance = subscription_to_subscription_instance(
                subscription = s,
                polling_services = polling_services,
                version = 11,
                subscription_parameters = s.params
            )

            response.subscription_instances.append(instance)

        return response


class SubscriptionRequest10Handler(BaseMessageHandler):
    supported_request_messages = [tm10.ManageFeedSubscriptionRequest]

    @classmethod
    def validate_request(cls, request):

        action = request.action

        if action not in ACT_TYPES_10:
            error_message = "The specified action was invalid"
        elif action == ACT_UNSUBSCRIBE and not request.subscription_id:
            error_message = 'Action "%s" requires a subscription id' % action
        else:
            error_message = None

        if error_message:
            raise StatusMessageException(ST_BAD_MESSAGE, message=error_message,
                    in_response_to=request.message_id)


    @classmethod
    def handle_message(cls, service, request):

        cls.validate_request(request)

        collection = retrieve_collection(service, request.feed_name, request.message_id)

        if request.subscription_id:
            subscription = service.get_subscription(request.subscription_id)
        else:
            subscription = None

        if subscription and subscription.collection_id != collection.id:
            details = {SD_ITEM: request.feed_name}
            raise StatusMessageException(ST_NOT_FOUND, status_details=details,
                    in_response_to=request.message_id)

        response = tm10.ManageFeedSubscriptionResponse(
            message_id = cls.generate_id(),
            in_response_to = request.message_id,
            feed_name = collection.name,
            message = service.subscription_message,
        )

        results = ACTIONS[request.action](service=service, request=request,
                collection=collection, subscription=subscription, version=10)

        if not isinstance(results, list):
            results = [results]

        polling_services = service.get_polling_services(collection)

        for s in results:
            instance = subscription_to_subscription_instance(
                subscription = s,
                polling_services = polling_services,
                version = 10
            )

            response.subscription_instances.append(instance)

        return response


class SubscriptionRequestHandler(BaseMessageHandler):

    supported_request_messages = [tm11.ManageCollectionSubscriptionRequest, tm10.ManageFeedSubscriptionRequest]

    @classmethod
    def handle_message(cls, service, request):

        if isinstance(request, tm10.ManageFeedSubscriptionRequest):
            return SubscriptionRequest10Handler.handle_message(service, request)
        elif isinstance(request, tm11.ManageCollectionSubscriptionRequest):
            return SubscriptionRequest11Handler.handle_message(service, request)
        else:
            raise_failure("TAXII Message not supported by message handler", request.message_id)






class AuthManager(object):

    def __init__(self, api):
        self.api = api

    def authenticate(self, username, password):
        return self.api.authenticate(username, password)

    def get_account(self, token):
        return self.api.get_account(token)

    def create_account(self, username, password):
        return self.api.create_account(username, password)




from ..signals import (
    CONTENT_BLOCK_CREATED, INBOX_MESSAGE_CREATED,
    SUBSCRIPTION_CREATED
)
from ..taxii.converters import blob_to_service_entity


class PersistenceManager(object):
    '''Manager responsible for persisting and retrieving data.

    Manager uses API instance ``api`` for basic data CRUD operations and
    provides additional logic on top.

    :param `opentaxii.persistence.api.OpenTAXIIPersistenceAPI` api:
        instance of persistence API class
    '''

    def __init__(self, api):
        self.api = api

    # Methods only used in the CLI scripts provided with OpenTAXII.

    def create_service(self, entity):
        '''Create a service.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.
        '''
        return self.api.create_service(entity)

    def attach_collection_to_services(self, collection_id, service_ids):
        '''Attach collection to the services.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.
        '''
        return self.api.attach_collection_to_services(collection_id, service_ids)

    def create_collection(self, entity):
        '''Create a collection.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.
        '''
        return self.api.create_collection(entity)

    def create_services_from_object(self, services_config):
        '''Create services from configuration object and persis them.

        NOTE: Additional data management method that is not used
        in TAXII server logic.
        '''

        for blob in services_config:
            self.create_service(blob_to_service_entity(blob))

    # =================================================================

    def get_services(self):
        '''Get configured services.

        Methods loads services entities via persistence API.

        :return: list of service entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.ServiceEntity`
        '''
        return self.api.get_services()

    def get_services_for_collection(self, collection):
        '''Get the services associated with a collection.

        :param `opentaxii.taxii.entities.CollectionEntity` collection:
            collection entity in question

        :return: list of service entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.ServiceEntity`
        '''

        return self.api.get_services(collection_id=collection.id)

    def get_collections(self, service_id):
        '''Get the collections associated with a service.

        :param str service_id: ID of the service in question

        :return: list of collection entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''

        return self.api.get_collections(service_id)

    def get_collection(self, name, service_id):
        '''Get a collection by name and service ID.

        According to TAXII spec collection name is unique per service instance.
        Method retrieves collection entity using collection name ``name`` and service ID
        ``service_id`` as a composite key.

        :param str name: a collection name
        :param str service_id: ID of a service

        :return: collection entity
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''
        return self.api.get_collection(name, service_id)


    def create_inbox_message(self, entity):
        '''Create an inbox message.

        Methods emits :py:const:`opentaxii.signals.INBOX_MESSAGE_CREATED` signal.

        :param `opentaxii.taxii.entities.InboxMessageEntity` entity:
            inbox message entity in question
        :return: updated inbox message entity
        :rtype: :py:class:`opentaxii.taxii.entities.InboxMessageEntity`
        '''

        created = self.api.create_inbox_message(entity)

        INBOX_MESSAGE_CREATED.send(self, inbox_message=created)

        return created

    def create_content(self, content, service_id=None, inbox_message=None,
            collections=None):
        '''Create a content block.

        Methods emits :py:const:`opentaxii.signals.CONTENT_BLOCK_CREATED` signal.

        :param `opentaxii.taxii.entities.ContentBlockEntity` entity:
                content block in question
        :param str service_id: ID of an inbox service via which content
                block was created
        :param `opentaxii.taxii.entities.InboxMessageEntity` inbox_message:
                inbox message that delivered the content block
        :param list collections: a list of destination collections as
                :py:class:`opentaxii.taxii.entities.CollectionEntity`
        :return: updated content block entity
        :rtype: :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        '''

        if inbox_message:
            if not inbox_message.id:
                inbox_message = self.api.create_inbox_message(inbox_message)
            content.inbox_message_id = inbox_message.id

        collections = collections or []
        collection_ids = [c.id for c in collections]
        content = self.api.create_content_block(content,
                collection_ids=collection_ids, service_id=service_id)

        CONTENT_BLOCK_CREATED.send(self, content_block=content,
                collection_ids=collection_ids, service_id=service_id)

        return content

    def get_content_blocks_count(self, collection_id, start_time=None, end_time=None,
            bindings=None):
        '''Get a count of the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`
            
        :return: content block count
        :rtype: int
        '''

        return self.api.get_content_blocks_count(
            collection_id = collection_id,
            start_time = start_time,
            end_time = end_time,
            bindings = bindings or [],
        )

    def get_content_blocks(self, collection_id, start_time=None, end_time=None,
            bindings=None, offset=0, limit=10):
        '''Get the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`
        :param int offset: result set offset
        :param int limit: result set max size
            
        :return: content blocks list
        :rtype: list of :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        '''

        return self.api.get_content_blocks(
            collection_id = collection_id,
            start_time = start_time,
            end_time = end_time,
            bindings = bindings or [],
            offset = offset,
            limit = limit,
        )

    def create_result_set(self, entity):
        '''Create a result set.

        :param `opentaxii.taxii.entities.ResultSetEntity` entity:
            result set entity in question
            
        :return: updated result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        '''
        return self.api.create_result_set(entity)

    def get_result_set(self, result_set_id):
        '''Get a result set entity by ID.

        :param str result_set_id: ID of a result set.

        :return: result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        '''
        return self.api.get_result_set(result_set_id)

    def create_subscription(self, entity):
        '''Create a subscription.

        Methods emits :py:const:`opentaxii.signals.SUBSCRIPTION_CREATED` signal.

        :param `opentaxii.taxii.entities.SubscriptionEntity` entity:
            subscription entity in question.

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''

        created = self.api.create_subscription(entity)

        SUBSCRIPTION_CREATED.send(self, subscription=created)

        return created

    def get_subscription(self, subscription_id):
        '''Get a subscription entity by ID.

        :param str subscription_id: ID of a subscription

        :return: subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        return self.api.get_subscription(subscription_id)

    def get_subscriptions(self, service_id):
        '''Get the subscriptions attached to/created via a service.

        :param str service_id: ID of a service

        :return: list of subscription entities
        :rtype: list of :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        return self.api.get_subscriptions(service_id=service_id)

    def update_subscription(self, subscription):
        '''Update a subscription status.

        :param `opentaxii.taxii.entities.SubscriptionEntity` subscription:
            subscription entity in question

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        return self.api.update_subscription(subscription)

